@@ -18,13 +18,15 @@
 
 align <- function(index,readfile1,readfile2=NULL,type="rna",input_format="gzFASTQ",output_format="BAM",output_file=paste(readfile1,"subread",output_format,sep="."),phredOffset=33,nsubreads=10,TH1=3,TH2=1,maxMismatches=3,unique=FALSE,nBestLocations=1,indels=5,complexIndels=FALSE,nTrim5=0,nTrim3=0,minFragLength=50,maxFragLength=600,PE_orientation="fr",nthreads=1,readGroupID=NULL,readGroup=NULL,keepReadOrder=FALSE,sortReadsByCoordinates=FALSE,color2base=FALSE,DP_GapOpenPenalty=-1,DP_GapExtPenalty=0,DP_MismatchPenalty=0,DP_MatchScore=2,detectSV=FALSE,useAnnotation=FALSE,annot.inbuilt="mm10",annot.ext=NULL,isGTF=FALSE,GTF.featureType="exon",GTF.attrType="gene_id",chrAliases=NULL)
 {
-  if(!.is.64bit.system()) cat("WARNING: your system seems to be 32-bit. Rsubread supports 32-bit sustems to a very limited level.\nIt is highly recommended to run Rsubread on a 64-bit system to avoid errors.\n\n")
+  if(!.is.64bit.system()) cat("WARNING: your system seems to be 32-bit. Rsubread supports 32-bit systems to a very limited level.\nIt is highly recommended to run Rsubread on a 64-bit system to avoid errors.\n\n")
 
   extra.params <- .retrieve.tmp.parameters()
   isBCLinput <- FALSE
   isScRNAFastqinput <- FALSE
+  isScRNABAMinput <- FALSE
   if("isBCLinput" %in% names(extra.params))isBCLinput <- extra.params[["isBCLinput"]]
   if("isScRNAFastqinput" %in% names(extra.params))isScRNAFastqinput <- extra.params[["isScRNAFastqinput"]]
+  if("isScRNABAMinput" %in% names(extra.params)) isScRNABAMinput <- extra.params[["isScRNABAMinput"]]
 
   .check_string_param(input_format,"input_format")
   .check_string_param(output_format,"output_format")
@@ -94,6 +96,9 @@ align <- function(index,readfile1,readfile2=NULL,type="rna",input_format="gzFAST
   if(isScRNAFastqinput)
     opt <- paste(opt,"--scRNA_FQinput",sep=.R_param_splitor)
 
+  if(isScRNABAMinput)
+    opt <- paste(opt,"--scRNA_BAMinput",sep=.R_param_splitor)
+
   if(!unique)
     opt <- paste(opt,"--multiMapping",sep=.R_param_splitor)
 
