@@ -23,6 +23,7 @@
 #include <stdarg.h>
 #include <math.h>
 #include <unistd.h>
+#include <pthread.h>
 
 
 #ifdef MACOS
@@ -58,6 +59,20 @@
 #include "gene-algorithms.h"
 #include "HelperFunctions.h"
 
+void sorting_LLU_array_exchange(void * varr, int i, int j){
+	srUInt_64 *arr = varr;
+	srUInt_64 t = arr[i];
+	arr[i]=arr[j];
+	arr[j]=t;
+}
+
+int sorting_LLU_array_compare(void * varr, int i, int j){
+	srUInt_64 *arr = varr;
+	if(arr[i]>arr[j])return 1;
+	if(arr[i]<arr[j])return -1;
+	return 0;
+}
+
 size_t get_sys_mem_info(char * keyword){
 	FILE * mfp = fopen("/proc/meminfo","r");
 	if(mfp==NULL) return -1;
@@ -1257,9 +1272,10 @@ int rebuild_command_line(char ** lineptr, int argc, char ** argv){
 
 	for(c = 0; c<argc;c++)
 	{
-		if(strlen(*lineptr) + 500 > linecap)
+		int cline = strlen(argv[c]);
+		if(strlen(*lineptr) + 100 + cline > linecap)
 		{
-			linecap *=2;
+			linecap += cline+500;
 			*lineptr = realloc(*lineptr, linecap);
 		}
 		sprintf((*lineptr) + strlen(*lineptr), "\"%s\" ", argv[c]);
@@ -2737,3 +2753,129 @@ int get_free_total_mem(size_t * total, size_t * free_mem){
 #endif
 #endif
 }
+
+void worker_master_mutex_init(worker_master_mutex_t * wmt, int all_workers){
+	memset(wmt,0,sizeof(worker_master_mutex_t));
+	wmt -> conds_worker_wait = malloc(sizeof(pthread_cond_t) * all_workers);
+	wmt -> mutexs_worker_wait = malloc(sizeof(pthread_mutex_t) * all_workers);
+	wmt -> mutex_with_master = calloc(sizeof(int), all_workers);
+	wmt -> worker_is_working = calloc(sizeof(int), all_workers);
+
+	wmt -> workers = all_workers;
+	int x1;
+	for(x1=0;x1<all_workers;x1++){
+		pthread_cond_init(&wmt -> conds_worker_wait [x1], NULL);
+		pthread_mutex_init(&wmt -> mutexs_worker_wait [x1], NULL);
+	}
+}
+
+void worker_thread_start(worker_master_mutex_t * wmt, int worker_id){
+	pthread_mutex_lock(&wmt->mutexs_worker_wait[worker_id]);
+	wmt -> mutex_with_master[worker_id] = 0;
+}
+
+void worker_master_mutex_destroy(worker_master_mutex_t * wmt){
+	int x1;
+	for(x1=0;x1<wmt -> workers;x1++){
+		pthread_mutex_destroy(&wmt -> mutexs_worker_wait [x1]);
+		pthread_cond_destroy(&wmt -> conds_worker_wait [x1]);
+	}
+	free( wmt -> worker_is_working);
+	free( wmt -> mutex_with_master);
+	free( wmt -> conds_worker_wait);
+	free( wmt -> mutexs_worker_wait);
+}
+
+int worker_wait_for_job(worker_master_mutex_t * wmt, int worker_id){
+	pthread_mutex_trylock(wmt->mutexs_worker_wait + worker_id);
+	wmt->worker_is_working[worker_id] = 0;
+	while(1){
+		pthread_cond_wait(&wmt->conds_worker_wait[worker_id], &wmt->mutexs_worker_wait[worker_id]);
+		if(wmt -> all_terminate)
+			pthread_mutex_unlock(&wmt->mutexs_worker_wait[worker_id]);
+		if(wmt -> mutex_with_master[worker_id] == 0)break;
+		// otherwise, it is a spurious wakeup
+	}
+	return wmt -> all_terminate;
+}
+
+void master_wait_for_job_done(worker_master_mutex_t * wmt, int worker_id){
+	if(!wmt -> mutex_with_master[worker_id] ){
+		while(1){
+			pthread_mutex_lock(&wmt->mutexs_worker_wait[worker_id]);
+			if(!wmt->worker_is_working[worker_id])break;
+			// In a rare situation, the worker hasn't been scheduled after the last notification.
+			// Namely, the master will notify twice the worker when the worker thread hasn't started at all. The second notification will cause no condition signaled.
+			// This test is to prevent such a situation.
+			pthread_mutex_unlock(&wmt->mutexs_worker_wait[worker_id]);
+			usleep(50);
+		}
+		if(wmt->worker_is_working[worker_id])
+			SUBREADprintf("ERROR 3: HOW CAN I HAVE THIS LOCK : %d?? TH_%d\n", wmt->worker_is_working[worker_id], worker_id);
+	}
+	wmt -> mutex_with_master[worker_id] = 1;
+}
+
+// master collects results between master_wait_for_job_done and master_notify_worker
+void master_notify_worker(worker_master_mutex_t * wmt, int worker_id){
+	if(!wmt -> mutex_with_master[worker_id])SUBREADprintf("ERROR 2: HOW CAN I NOT HAVE THE LOCK : %d ; TERM=%d\n", worker_id, wmt -> all_terminate);
+
+	wmt -> worker_is_working[worker_id] = 1;
+	wmt -> mutex_with_master[worker_id] = 0;
+	pthread_cond_signal(&wmt->conds_worker_wait[worker_id]);
+	pthread_mutex_unlock(&wmt->mutexs_worker_wait[worker_id]);
+}
+
+// this function must be called when the master thread has all the worker locks in control.
+// ie all the workers should be in the "wait_for_job" status.
+void terminate_workers(worker_master_mutex_t * wmt){
+	int x1;
+	wmt -> all_terminate = 1;
+	for(x1=0;x1<wmt -> workers;x1++)
+		master_notify_worker(wmt,x1);
+}
+
+void *windows_memmem(const void *haystack_start, size_t haystack_len, const void *needle_start, size_t needle_len)
+{
+
+    const unsigned char *haystack = (const unsigned char *) haystack_start;
+    const unsigned char *needle = (const unsigned char *) needle_start;
+    const unsigned char *h = NULL;
+    const unsigned char *n = NULL;
+    size_t x = needle_len;
+
+    /* The first occurrence of the empty string is deemed to occur at
+ *     the beginning of the string.  */
+    if (needle_len == 0)
+        return (void *) haystack_start;
+
+    /* Sanity check, otherwise the loop might search through the whole
+ *         memory.  */
+     if (haystack_len < needle_len)
+       return NULL;
+
+    for (; *haystack && haystack_len--; haystack++) {
+
+        x = needle_len;
+        n = needle;
+        h = haystack;
+
+        if (haystack_len < needle_len)
+            break;
+
+        if ((*haystack != *needle) || ( *haystack + needle_len != *needle + needle_len))
+            continue;
+
+        for (; x ; h++ , n++) {
+            x--;
+
+            if (*h != *n) 
+                break;
+
+           if (x == 0)
+            return (void *)haystack;
+        }
+    }
+
+    return NULL;
+}
