@@ -138,7 +138,7 @@ typedef struct{
 	int fast_run;
 
 	// input_scheme
-	char first_read_file[MAX_FILE_NAME_LENGTH];
+	char first_read_file[MAX_FILE_NAME_LENGTH * MAX_SCRNA_FASTQ_FILES * 3];
 	char second_read_file[MAX_FILE_NAME_LENGTH];
 	char exon_annotation_file[MAX_FILE_NAME_LENGTH];
 	char exon_annotation_file_screen_out[MAX_FILE_NAME_LENGTH];
@@ -148,6 +148,7 @@ typedef struct{
 	char exon_annotation_feature_name_column[MAX_READ_NAME_LEN];
 
 
+	int do_remove_neighbour_for_scRNA;
 	int scRNA_input_mode;
 	short read_trim_5;
 	short read_trim_3;
@@ -411,7 +412,10 @@ typedef struct
 // THREE_TOP_UPDATE dictates the number of top-votes.
 
 typedef struct {
-	int is_vote_t_item;
+	union{
+		int is_vote_t_item;
+		int is_exon_item;
+	};
 	int item_index_i;
 	int item_index_j;
 	unsigned int mapping_position;
@@ -694,4 +698,5 @@ int exec_cmd(char * cmd, char * outstr, int out_limit);
 int is_pos_in_annotated_exon_regions(global_context_t * global_context, unsigned int pos);
 char * get_sam_chro_name_from_alias(HashTable * tab, char * anno_chro);
 void subread_rebuild_cmd(int argc, char ** argv, global_context_t * global_context);
+int get_soft_clipping_length(char* CIGAR);
 #endif
