@@ -32,9 +32,11 @@
 #include <ctype.h>
 #include "subread.h"
 #include "core.h"
+#include "HelperFunctions.h"
 #include "gene-algorithms.h"
 #include "sambam-file.h"
 #include "input-files.h"
+#define Z_DEFAULT_MEM_LEVEL  8
 
 int SamBam_fetch_next_chunk(SamBam_FILE *fp)
 {
@@ -58,7 +60,7 @@ int SamBam_fetch_next_chunk(SamBam_FILE *fp)
 			int nchunk = 0;
 			ret = PBam_get_next_zchunk(fp -> os_file, in_buff, 65536, & real_len);
 			if(ret > 0)
-				nchunk = SamBam_unzip(fp -> input_binary_stream_buffer + fp->input_binary_stream_write_ptr - fp -> input_binary_stream_read_ptr + have , in_buff , ret);
+				nchunk = SamBam_unzip(fp -> input_binary_stream_buffer + fp->input_binary_stream_write_ptr - fp -> input_binary_stream_read_ptr + have, 65536 , in_buff , ret, 0);
 			else if(ret == -2){
 				SUBREADputs("ERROR: BAM format is broken.");
 				return -2;
@@ -449,10 +451,12 @@ int PBam_get_next_zchunk(FILE * bam_fp, char * buffer, int buffer_length, unsign
 		int CDATA_READING = min(CDATA_LEN, buffer_length);
 		rrtv = fread(buffer, 1, CDATA_READING, bam_fp);
 		if(rrtv < CDATA_READING) return -1;
-		if(CDATA_READING<CDATA_LEN)
-			fseeko(bam_fp, CDATA_LEN-CDATA_READING, SEEK_CUR);
+		if(CDATA_READING<CDATA_LEN){
+			SUBREADputs("ERROR: buffer insufficient");
+			return -1;
+		}
 		fseeko(bam_fp, 4, SEEK_CUR);
-		rlen = fread(&real_len, 4, 1, bam_fp);
+		rlen = fread(&real_len, 4, 1, bam_fp);//Input SIZE (length of uncompressed data);uint32_t
 		if(rlen < 1) is_file_broken = 1;
 
 	//	SUBREADprintf("read_data=%u\n", CDATA_LEN);
@@ -1090,7 +1094,7 @@ int test_bamview(int argc, char ** argv)
 	return 0;
 }
 
-int SamBam_writer_create(SamBam_Writer * writer, char * BAM_fname, int threads, int sort_reads_by_coord, char * tmpfname)
+int SamBam_writer_create(SamBam_Writer * writer, char * BAM_fname, int threads, int sort_reads_by_coord, int is_temp_BAM_file, char * tmpfname)
 {
 	memset(writer, 0, sizeof(SamBam_Writer));
 
@@ -1101,6 +1105,7 @@ int SamBam_writer_create(SamBam_Writer * writer, char * BAM_fname, int threads,
 			char tname[MAX_FILE_NAME_LENGTH];
 			sprintf(tname, "%s.bai", BAM_fname);
 			writer -> BAI_fp = f_subr_open(tname, "wb");
+			worker_master_mutex_init(&writer->sorted_notifier, threads);
 		}
 		if(!writer -> bam_fp) return -1;
 	}
@@ -1111,7 +1116,7 @@ int SamBam_writer_create(SamBam_Writer * writer, char * BAM_fname, int threads,
 
 	writer -> threads = threads;
 	writer -> sort_reads_by_coord = sort_reads_by_coord;
-	writer -> fastest_compression = 0;
+	writer -> fastest_compression = is_temp_BAM_file;
 	writer -> compressed_chunk_buffer = malloc(70000); 
 	writer -> chunk_buffer = malloc(70000); 
 	writer -> chunk_buffer_max_size = 70000;
@@ -1194,13 +1199,19 @@ void SamBam_writer_add_chunk(SamBam_Writer * writer, int thread_no)
 	char * this_buffer = thread_no < 0 ? writer ->chunk_buffer: writer -> threads_chunk_buffer[thread_no];
 	char * this_compressed_chunk_buffer = thread_no < 0 ? writer ->compressed_chunk_buffer : writer -> threads_chunk_buffer_compressed[thread_no];
 
-	//if(thread_no>=0)SUBREADprintf("MTBAM : WRITE THR_%d ; LEN=%d\n", thread_no, *this_buffer_used);
+	if(*this_buffer_used < 1){
+		// magic block with 0 byte data(EOF)
+		subread_lock_occupy(&writer -> thread_bam_lock);
+		fwrite( "\x1f\x8b\x08\x04\x00\x00\x00\x00\x00\xff\x06\x00\x42\x43\x02\x00\x1b\x00\x03\x00\x00\x00\x00\x00\x00\x00\x00\x00",1,28,writer -> bam_fp);
+		writer -> current_BAM_pos = ftello(writer -> bam_fp);
+		subread_lock_release(&writer -> thread_bam_lock);
+		return;
+	}
 
 	this_stream -> avail_out = 70000;
 	this_stream -> avail_in = (* this_buffer_used);
 	CRC32 = SamBam_CRC32(this_buffer , * this_buffer_used);
 
-	int Z_DEFAULT_MEM_LEVEL = 8;
 	this_stream -> zalloc = Z_NULL;
 	this_stream -> zfree = Z_NULL;
 	this_stream -> opaque = Z_NULL;
@@ -1215,7 +1226,6 @@ void SamBam_writer_add_chunk(SamBam_Writer * writer, int thread_no)
 	deflateEnd(this_stream);
 
 	compressed_size = 70000 - this_stream -> avail_out;
-	//printf("ADDED BLOCK=%d; LEN=%d; S=%s\n", compressed_size, writer ->chunk_buffer_used,  writer ->chunk_buffer);
 
 	subread_lock_occupy(&writer -> thread_bam_lock);
 	SamBam_writer_chunk_header(writer, compressed_size);
@@ -1305,8 +1315,10 @@ int SamBam_writer_close(SamBam_Writer * writer)
 			SamBam_writer_write_header(writer);
 	} else SamBam_writer_finalise_one_thread(writer);
 
-	if(writer -> sort_reads_by_coord)
+	if(writer -> sort_reads_by_coord){
 		SamBam_writer_sort_bins_to_BAM(writer);
+		worker_master_mutex_destroy(&writer->sorted_notifier);
+	}
 	
 	writer -> chunk_buffer_used = 0;
 	SamBam_writer_add_chunk(writer, -1);
@@ -1623,67 +1635,6 @@ char * duplicate_TAB_record_field(char * rline, int fno, int to_end){
 	return ret;
 }
 
-int SamBam_writer_add_read_fqs_scRNA(gzFile *outfps, char * bambin){
-	int reclen=0;
-	gzFile outR1fp = outfps[0];
-	gzFile outI1fp = outfps[1];
-	gzFile outR2fp = outfps[2];
-
-	memcpy(&reclen, bambin,4);
-	int flag = 0, l_seq = 0, l_read_name = 0, n_cigar_ops = 0;
-	memcpy(&l_read_name, bambin+12,1);
-	memcpy(&n_cigar_ops, bambin+16,2);
-	memcpy(&flag, bambin+18,2);
-	memcpy(&l_seq, bambin+20,4);
-
-	int x1=0;
-
-	gzwrite(outR2fp,"@",1);
-	gzwrite(outR1fp,"@",1);
-	gzwrite(outI1fp,"@",1);
-	char * readname = bambin+36;
-	gzwrite(outR1fp,readname, 12);
-	gzwrite(outR2fp,readname, 12);
-	gzwrite(outI1fp,readname, 12);
-	gzwrite(outR1fp,"\n",1);
-	gzwrite(outR2fp,"\n",1);
-	gzwrite(outI1fp,"\n",1);
-
-	char * R1seq = bambin+36+13;
-	int R1len = 0;
-	for(R1len=0; R1seq[R1len] && R1seq[R1len]!='|' ;R1len++);
-	char * R1qual = R1seq + R1len + 1;
-	gzwrite(outR1fp,R1seq, R1len);
-	gzwrite(outR1fp,"\n+\n",3);
-	gzwrite(outR1fp,R1qual, R1len);
-	gzwrite(outR1fp,"\n",1);
-
-	char * I1seq = R1qual + R1len + 1;
-	int I1len = 0;
-	for(I1len=0; I1seq[I1len] && I1seq[I1len]!='|' ;I1len++);
-	char * I1qual = I1seq + I1len + 1;
-	gzwrite(outI1fp,I1seq, I1len);
-	gzwrite(outI1fp,"\n+\n",3);
-	gzwrite(outI1fp,I1qual, I1len);
-	gzwrite(outI1fp,"\n",1);
-
-	char oseq[l_seq+1];
-	int seqbase = 36+l_read_name+n_cigar_ops*4;
-	for(x1=0; x1<l_seq;x1++)oseq[x1]="=ACMGRSVTWYHKDBN"[(bambin[ seqbase+(x1/2)] >> (((x1%2)?0:1) *4) )&0xf];
-	oseq[l_seq]=0;
-	if(flag & 16) reverse_read(oseq,l_seq, GENE_SPACE_BASE);
-	gzwrite(outR2fp, oseq, l_seq);
-	gzwrite(outR2fp,"\n+\n",3);
-
-	seqbase = 36+l_read_name+n_cigar_ops*4 + ( l_seq+1 )/2;
-	for(x1=0; x1<l_seq;x1++)oseq[x1]=33+bambin[ seqbase+x1];
-	if(flag & 16)reverse_quality(oseq, l_seq);
-	oseq[l_seq]=0;
-	gzwrite(outR2fp, oseq, l_seq);
-	gzwrite(outR2fp,"\n",1);
-	return 0;
-}
-
 int SamBam_writer_add_read_bin(SamBam_Writer * writer, int thread_no, char * rbin, int committable){
 	char * this_chunk_buffer=NULL;
 	srInt_64 *this_chunk_buffer_used=NULL;
@@ -1696,7 +1647,7 @@ int SamBam_writer_add_read_bin(SamBam_Writer * writer, int thread_no, char * rbi
 		this_chunk_buffer_used = writer -> threads_chunk_buffer_used + thread_no;
 	}else{
 		if(writer -> sort_reads_by_coord && writer -> chunk_buffer_max_size < writer -> chunk_buffer_used + 12000){
-			//SUBREADprintf("REALLOCATE MEM : %d -> %d\n", writer -> chunk_buffer_max_size,  writer -> chunk_buffer_max_size * 7/4);
+			//SUBREADprintf("0THR_REALLOCATE MEM : %d -> %d\n", writer -> chunk_buffer_max_size,  writer -> chunk_buffer_max_size * 7/4);
 			writer -> chunk_buffer_max_size = writer -> chunk_buffer_max_size * 7/4;
 			writer -> chunk_buffer = realloc(writer -> chunk_buffer, writer -> chunk_buffer_max_size);
 		}
@@ -1845,9 +1796,8 @@ int SamBam_writer_add_read(SamBam_Writer * writer, int thread_no, char * read_na
 	return 0;
 }
 
-int SamBam_unzip(char * out , char * in , int inlen)
+int SamBam_unzip(char * out, int outlen , char * in , int inlen, int sync_only)
 {
-	#define unzip_out_max_len 65537
 	z_stream strm;
 	strm.zalloc = Z_NULL;
 	strm.zfree = Z_NULL;
@@ -1861,16 +1811,17 @@ int SamBam_unzip(char * out , char * in , int inlen)
 	strm.avail_in = (unsigned int)inlen;
 	strm.next_in = (unsigned char *)in;
 
-	strm.avail_out = unzip_out_max_len;
+	strm.avail_out = outlen;
 	strm.next_out = (unsigned char *)out;
-	ret = inflate(&strm, Z_FINISH);
-	if(ret != Z_STREAM_END)
+	ret = inflate(&strm, sync_only?Z_SYNC_FLUSH:Z_FINISH);
+	if(ret != Z_STREAM_END && ret!=0)
 	{
 		inflateEnd(&strm);
 		SUBREADprintf("DATA ERROR! code=%d\n", ret);
+		assert(0);
 		return -1;
 	}
-	int have = unzip_out_max_len - strm.avail_out;
+	int have = outlen - strm.avail_out;
 
 	inflateEnd(&strm);
 	//SUBREADprintf("DECOMPRESS GENERATED=%d\n", have);
@@ -1888,10 +1839,6 @@ int SamBam_writer_sort_buff_one_compare(void * Lbin, void * Rbin, ArrayList * me
 // return total reads in bin.
 int SamBam_writer_sort_buff_one_write(SamBam_Writer * writer, char * bin, int binlen, int thread_id){
 	int bin_cursor = 0;
-	if(binlen<1){
-		SUBREADprintf("WONE : BINLEN=%d, TH=%d\n", binlen, thread_id);
-		assert(binlen >0);
-	}
 
 	ArrayList* sort_linear_pos = ArrayListCreate(1000000);
 	ArrayListSetDeallocationFunction(sort_linear_pos,  free);
@@ -1913,7 +1860,8 @@ int SamBam_writer_sort_buff_one_write(SamBam_Writer * writer, char * bin, int bi
 	}
 	ArrayListSort(sort_linear_pos, SamBam_writer_sort_buff_one_compare);
 
-	char * nbin = malloc(binlen);
+	char * nbin = NULL;
+	if(binlen >0 && binlen < 0x7fffffff)nbin=malloc(binlen);
 	int nb_cursor = 0, xx, wlen=0;
 	for(xx=0; xx<ii_reads;xx++){
 		int * binpos = ArrayListGet(sort_linear_pos, xx);
@@ -1933,11 +1881,11 @@ int SamBam_writer_sort_buff_one_write(SamBam_Writer * writer, char * bin, int bi
 	if(writer -> threads>1) subread_lock_release(&writer -> thread_bam_lock);
 	FILE * tofp  = fopen(tmpfname, "wb");
 	if(tofp){
-		wlen = fwrite(nbin, binlen,1, tofp);
+		if(binlen>0)wlen = fwrite(nbin, binlen,1, tofp);
 		fclose(tofp);
 	}
 	free(nbin);
-	if(wlen < 1) {
+	if(wlen < 1 && binlen>0) {
 		SUBREADprintf("ERROR: no space (%d bytes) in the temp directory (%s).\nThe program cannot run properly.\n", binlen, tmpfname);
 		writer ->is_internal_error = 1;
 		return -1;
@@ -1946,37 +1894,31 @@ int SamBam_writer_sort_buff_one_write(SamBam_Writer * writer, char * bin, int bi
 }
 
 unsigned long long SamBam_writer_sort_bins_to_BAM_FP_pos(FILE * fp){
-	int blklen = 0;
+	int intvals[3];
 	unsigned long long ret = 0;
-	int rlen = fread(&blklen, 4,1,fp);
-	if(rlen>0 && blklen<10000){
-		int chro_no=0, poss=0;
-		rlen = fread(&chro_no, 4,1,fp);
-		if(rlen < 1) return SUBREAD_MAX_ULONGLONG;
-		rlen = fread(&poss, 4,1,fp);
-		if(rlen < 1) return SUBREAD_MAX_ULONGLONG;
+	int rlen = fread(intvals, sizeof(int),3,fp);
+	if(rlen>0 && intvals[0]<10000){
+		int chro_no, poss;
+		chro_no=intvals[1];
+		poss=intvals[2];
 		ret =((1LLU*chro_no) << 32)| poss;
 		if(ret == SUBREAD_MAX_ULONGLONG) ret = ret- 10 ;
-
-		fseek(fp,-12, SEEK_CUR);
-		//SUBREADprintf("READPOS AT %ld BLK=%d V=%llu\n", blklen, ftell(fp), ret);
+		fseek(fp, -12, SEEK_CUR);
 		return ret;
-	}else if(rlen < 1) return SUBREAD_MAX_ULONGLONG;
-	assert(rlen >0 && blklen<10000);
+	}
 	return SUBREAD_MAX_ULONGLONG;
 }
 
 
-int SamBam_writer_calc_cigar_span(char * bin, int blen){
-	int cops = 0, rname_len = 0, ret = 0;
-	memcpy(&cops, bin+12, 4);
-	memcpy(&rname_len, bin+8, 4);
-	rname_len = rname_len & 0xff;
-	cops = cops & 0xffff;
+int SamBam_writer_calc_cigar_span(char * bin){
+	int cops, rname_len, ret = 0;
+	cops = *(short *)(bin+12);
+	rname_len = bin[8];
 	int ii;
+	char * binbase = bin+32+rname_len;
 	for(ii = 0; ii < cops ; ii++){
 		unsigned int copt = 0;
-		memcpy(&copt, bin+32+rname_len+4*ii, 4);
+		copt = *(int*)(binbase+4*ii);
 		int copt_char = copt & 0xf;
 		unsigned int copt_len = copt >> 4;
 		if(copt_char == 0 || copt_char == 2 || copt_char == 3 || copt_char == 7 || copt_char == 8) ret += copt_len;
@@ -1985,7 +1927,7 @@ int SamBam_writer_calc_cigar_span(char * bin, int blen){
 	return ret;
 }
 
-#define MAX_ALLOWED_GAP_IN_BAI_CHUNK 0x10000 
+#define MAX_ALLOWED_GAP_IN_BAI_CHUNK 5 // the number of blocks, not the file positions.
 
 void SamBam_writer_sort_bins_to_BAM_test_bins(SamBam_Writer * writer, HashTable * bin_tab, ArrayList * bin_list, ArrayList * win16k_list, int block_len, void ***last_chunk_ptr, int chro_no){
 	int inbin_pos = writer -> chunk_buffer_used - block_len ; // point to the byte AFTER "block_len" int.
@@ -1995,76 +1937,143 @@ void SamBam_writer_sort_bins_to_BAM_test_bins(SamBam_Writer * writer, HashTable
 	memcpy(&bin_mq_nl, writer -> chunk_buffer + inbin_pos + 8,4);
 	binno = bin_mq_nl>>16;
 
-	int cigar_span = SamBam_writer_calc_cigar_span(writer -> chunk_buffer + inbin_pos, block_len);
-
+	int cigar_span = SamBam_writer_calc_cigar_span(writer -> chunk_buffer + inbin_pos);
 	int this_w16_no = (pos + cigar_span) >>14;	// WIN is calculated on 0-based pos.
-	unsigned long long this_Vpos = writer -> current_BAM_pos<<16 | (inbin_pos-4);
-
-	if(0){
-		if(0 && binno == 10540 && chro_no == 0)
-				SUBREADprintf("IN_BIN_10540/chr0 : %s\n", writer -> chunk_buffer + inbin_pos + 32);
-
-		if(FIXLENstrcmp("V0112_0155:7:1302:1337:112281\nV0112_0155:7:1107:11281:15430\nV0112_0155:7:1103:16366:84183\nV0112_0155:7:1205:9418:2969", writer -> chunk_buffer + inbin_pos + 32) == 0){
-			SUBREADprintf("WRITE %s\tBINNO=%d\n", writer -> chunk_buffer + inbin_pos + 32, binno);
-			//SUBREADprintf("WRITE %s AT  %lld, %d    BINNO=%d   WINNO=%d\n", writer -> chunk_buffer + inbin_pos + 32, this_Vpos>>16,(int)(this_Vpos&0xffff), binno, this_w16_no);
-		}
-	}
+	unsigned long long this_Vpos = writer -> this_bam_block_no<<16 | (inbin_pos-4);
 
+	// if this read is after the maximum coordinate in the win16k list: all elements before last one and this one starts at this read.
 	if(this_w16_no >=win16k_list->numOfElements){
 		int bbi;
 		for(bbi = win16k_list->numOfElements; bbi <=this_w16_no; bbi++)
 			ArrayListPush(win16k_list, NULL+ this_Vpos);
 	}
 
-	ArrayList * this_bin_chinks = HashTableGet(bin_tab, NULL+binno+1);
-	if(NULL == this_bin_chinks){
-		this_bin_chinks = ArrayListCreate(5);
-		HashTablePut(bin_tab, NULL+binno+1, this_bin_chinks);
+	// a read only belongs to ONE bin which is the smallest bin that can FULLY cover the full length of this read.
+	ArrayList * this_bin_chunks = HashTableGet(bin_tab, NULL+binno+1);
+	if(NULL == this_bin_chunks){
+		this_bin_chunks = ArrayListCreate(5);
+		HashTablePut(bin_tab, NULL+binno+1, this_bin_chunks);
 		ArrayListPush(bin_list, NULL+binno);
 	}
 
 	int found = 0;
-	if(this_bin_chinks -> numOfElements > 0){
-		//SUBREADprintf("RESLOCS : %ld == %ld\n", this_Vpos, this_bin_chinks -> elementList [ this_bin_chinks -> numOfElements - 1 ] - NULL);
+	// a bin is not necessarily continuous. Say, a top-level bin only contains a few reads (most reads a in low-level bins), but their locations are everywhere.
+	if(this_bin_chunks -> numOfElements > 0){
+		//SUBREADprintf("RESLOCS : %ld == %ld\n", this_Vpos, this_bin_chunks -> elementList [ this_bin_chunks -> numOfElements - 1 ] - NULL);
 		long long diff = this_Vpos >>16;
-		diff -=(this_bin_chinks -> elementList [ this_bin_chinks -> numOfElements - 1 ] - NULL)>>16;
+		diff -=(this_bin_chunks -> elementList [ this_bin_chunks -> numOfElements - 1] - NULL)>>16;
 		if(diff < MAX_ALLOWED_GAP_IN_BAI_CHUNK){
-			this_bin_chinks -> elementList [ this_bin_chinks -> numOfElements -1] = NULL+this_Vpos + block_len+4;
-			*last_chunk_ptr = this_bin_chinks -> elementList + this_bin_chinks -> numOfElements -1;
+			this_bin_chunks -> elementList [ this_bin_chunks -> numOfElements -1] = NULL+this_Vpos + block_len+4;
+			*last_chunk_ptr = this_bin_chunks -> elementList + this_bin_chunks -> numOfElements -1;
 			found = 1;
 		}
 	}
+
+	// if the last chunk in this bin isn't good to be extended (too far from the file location of the new read), a new chunk is created.
 	if(!found){
-		ArrayListPush(this_bin_chinks, NULL + this_Vpos);
-		ArrayListPush(this_bin_chinks, NULL + this_Vpos + block_len+4);
-		*last_chunk_ptr = this_bin_chinks -> elementList + this_bin_chinks -> numOfElements -1;
+		ArrayListPush(this_bin_chunks, NULL + this_Vpos);
+		ArrayListPush(this_bin_chunks, NULL + this_Vpos + block_len+4);
+		*last_chunk_ptr = this_bin_chunks -> elementList + this_bin_chunks -> numOfElements -1;
 	}
 }
 
+void * SamBam_writer_sorted_compress(void * vptr0){
+	void ** vptr = vptr0;
+	SamBam_Writer * writer = vptr[0];
+	int this_thread_no = vptr[1]-NULL;
+	subread_lock_t * initlock = vptr[2];
+
+	worker_thread_start(&writer->sorted_notifier, this_thread_no);
+	subread_lock_release(initlock);
+	free(vptr0);
+	struct SamBam_sorted_compressor_st * me = writer -> writer_threads + this_thread_no;
+
+	while(1){
+		int termed = worker_wait_for_job(&writer->sorted_notifier, this_thread_no);
+		if(termed)break;
+
+		me->CRC32_plain = SamBam_CRC32(me->plain_text,  me->text_size);
+		me->strm.next_in = (unsigned char*)me->plain_text;
+		me->strm.avail_in = me->text_size;
+		me->strm.next_out = (unsigned char*)me->zipped_bin;
+		me->strm.avail_out = 70000;
+		int deret = deflate(&me->strm,Z_FINISH);
+		if(deret >=0){
+			deflateReset(&me->strm);
+			me->bin_size = 70000 - me->strm.avail_out;
+			me->last_job_done = 1;
+		}else{
+			SUBREADprintf("Error: cannot compress BAM block #%d , which is %llu, had %d => 70000 [ %d ] bytes , return = %d\n", this_thread_no, me -> bam_block_no, me->text_size, me->bin_size, deret);
+		}
+	}
+	return NULL;
+}
+
+void SamBam_thread_wait_merge_write(SamBam_Writer * writer, int thread_no){
+	master_wait_for_job_done(&writer -> sorted_notifier, thread_no);
+
+	if(writer -> writer_threads[thread_no].last_job_done){
+		srInt_64 fpos = ftello(writer -> bam_fp);
+		HashTablePut(writer -> block_no_p1_to_vpos_tab, NULL+1+writer -> writer_threads[thread_no].bam_block_no, NULL+fpos);
+
+		SamBam_writer_chunk_header(writer, writer -> writer_threads[thread_no].bin_size);
+		int rlen = fwrite( writer -> writer_threads[thread_no].zipped_bin ,1, writer -> writer_threads[thread_no].bin_size, writer -> bam_fp);
+		if(rlen != writer -> writer_threads[thread_no].bin_size){
+			SUBREADprintf("ERROR: cannot write output files.\n");
+			assert(0);
+		}
+		rlen = fwrite(&writer -> writer_threads[thread_no].CRC32_plain , 4, 1, writer -> bam_fp);
+		rlen = fwrite(&writer -> writer_threads[thread_no].text_size , 4, 1, writer -> bam_fp);
+
+		writer -> writer_threads[thread_no].text_size = 0;
+		writer -> writer_threads[thread_no].bin_size = 0;
+		writer -> writer_threads[thread_no].bam_block_no = -1;
+		writer -> writer_threads[thread_no].last_job_done = 0;
+	}
+}
+
+void SamBam_thread_set_new_job(SamBam_Writer * writer, int thread_no){
+	memcpy(writer -> writer_threads[thread_no].plain_text, writer ->chunk_buffer, writer->chunk_buffer_used);
+	writer -> writer_threads[thread_no].text_size = writer->chunk_buffer_used;
+	writer -> writer_threads[thread_no].bam_block_no = writer -> this_bam_block_no;
+	writer -> chunk_buffer_used = 0;
+	master_notify_worker(&writer -> sorted_notifier, thread_no);
+}
+
+void SamBam_writer_submit_sorted_compressing_task(SamBam_Writer * writer){
+	SamBam_thread_wait_merge_write(writer, writer -> sorted_compress_this_thread_no);
+	SamBam_thread_set_new_job(writer, writer -> sorted_compress_this_thread_no);
+
+	writer -> sorted_compress_this_thread_no++;
+	if(writer -> sorted_compress_this_thread_no == writer -> threads) writer -> sorted_compress_this_thread_no=0;
+	writer -> this_bam_block_no ++;
+}
+
 void SamBam_writer_sort_bins_to_BAM_write_1R(SamBam_Writer * writer, FILE * fp, HashTable * bin_tab, ArrayList * bin_list, ArrayList * win16k_list, int chro_no){
 	int block_len=0;
 	int rlen = fread(&block_len, 4,1,fp);
 	//SUBREADprintf("WBIN=%d\n", block_len);
 	if(rlen<1 || block_len >= 10000){
-		SUBREADprintf("ERROR: sorted bin files are brolen. RLEN=%d , BLKLEN=%d\n", rlen, block_len);
+		SUBREADprintf("ERROR: sorted bin files are broken. RLEN=%d , BLKLEN=%d\n", rlen, block_len);
 		assert(rlen >=1);
 	}
 
 	memcpy(writer -> chunk_buffer + writer -> chunk_buffer_used, &block_len, 4);
-	writer -> chunk_buffer_used += 4;
+	writer ->chunk_buffer_used += 4;
 	rlen = fread(writer -> chunk_buffer + writer -> chunk_buffer_used, 1, block_len ,fp);
 	if(rlen < block_len){
-		SUBREADprintf("ERROR: sorted bin files are brolen.\n");
+		SUBREADprintf("ERROR: sorted bin files are broken.\n");
 		assert(rlen >=block_len);
 	}
-	writer -> chunk_buffer_used += rlen;
+	writer -> chunk_buffer_used+= rlen;
 
 	void ** last_chunk_ptr = NULL;
 	SamBam_writer_sort_bins_to_BAM_test_bins(writer, bin_tab, bin_list, win16k_list, block_len, &last_chunk_ptr, chro_no);
 
-	if(writer -> chunk_buffer_used >55000)
-		SamBam_writer_add_chunk(writer, -1);
-	//if(last_chunk_ptr) (*last_chunk_ptr) = NULL + (writer -> current_BAM_pos << 16);
+	if(writer -> chunk_buffer_used>55000){
+		assert(writer -> chunk_buffer_used< 65000);
+		SamBam_writer_submit_sorted_compressing_task(writer);
+	}
 }
 
 #define SAMBAM_MERGE_MAX_FPS (350) 
@@ -2277,6 +2286,64 @@ void SamBam_writer_optimize_bins(HashTable *bin_tab, ArrayList *bin_arr, HashTab
 	}
 }
 
+void SamBam_write_sorted_thread_collect(SamBam_Writer * writer){
+	int thread_i;
+	if(writer -> chunk_buffer_used>0) SamBam_writer_submit_sorted_compressing_task(writer);
+	for(thread_i = 0; thread_i < writer -> threads; thread_i++){
+		//SUBREADprintf("FINISHED %d-th CHRO, MERGING THR %d\n", old_chro_no, writer->sorted_compress_this_thread_no );
+		SamBam_thread_wait_merge_write(writer, writer->sorted_compress_this_thread_no);
+		writer->sorted_compress_this_thread_no ++;
+		if(writer->sorted_compress_this_thread_no == writer -> threads) writer->sorted_compress_this_thread_no = 0;
+	}
+}
+
+#define SAMBAM_Block2Vpos(v) (v)=((HashTableGet(writer -> block_no_p1_to_vpos_tab, NULL+(v>>16)+1)-NULL )<<16 )|(v & 0xffff);
+void SamBam_write_BAI_for_1chr(SamBam_Writer * writer, HashTable ** bin_chunks_table, ArrayList ** bins_list, ArrayList ** win16k_list){
+	SamBam_write_sorted_thread_collect(writer);
+
+	if(1){
+		HashTable * new_bin_tab = NULL;
+		ArrayList * new_bin_list = NULL;
+		SamBam_writer_optimize_bins(*bin_chunks_table ,*bins_list,&new_bin_tab,&new_bin_list);
+		*bin_chunks_table = new_bin_tab;
+		*bins_list = new_bin_list;
+	}
+
+	int n_bins = (*bins_list)->numOfElements;
+	int win16_nos = (*win16k_list)->numOfElements;
+	//SUBREADprintf("WIN_ITEMS=%d\t\t\tOLD_CHRO=%d  NEW_CHRO=%d  CURR_BATCH=%d\n", win16_nos, old_chro_no, chro_no , current_min_batch);
+
+	fwrite(&n_bins, 4, 1, writer -> BAI_fp);
+	int bini;
+	for(bini = 0; bini < n_bins; bini ++){
+		int bin_no = ArrayListGet(*bins_list, bini)-NULL;
+		ArrayList * chunks_list = HashTableGet(*bin_chunks_table, NULL+ bin_no+1);
+
+		int n_chunks = chunks_list -> numOfElements/2;
+		fwrite(&bin_no, 4, 1, writer -> BAI_fp);
+		fwrite(&n_chunks, 4, 1, writer -> BAI_fp);
+		int chunk_i;
+		for(chunk_i = 0; chunk_i < 2* n_chunks; chunk_i +=2){
+			long long int Voff_st = ArrayListGet(chunks_list, chunk_i) - NULL;
+			long long int Voff_en = ArrayListGet(chunks_list, chunk_i+1) - NULL;
+			SAMBAM_Block2Vpos(Voff_st);
+			SAMBAM_Block2Vpos(Voff_en);
+			fwrite(&Voff_st, 8, 1, writer -> BAI_fp);
+			fwrite(&Voff_en, 8, 1, writer -> BAI_fp);
+		}
+	}
+
+	fwrite(&win16_nos, 4, 1, writer -> BAI_fp);
+	for(bini = 0; bini < win16_nos; bini++){
+		long long int Voff = ArrayListGet(*win16k_list , bini) - NULL;
+		SAMBAM_Block2Vpos(Voff);
+		fwrite(&Voff, 8, 1, writer -> BAI_fp);
+	}
+
+}
+
+#define SAMBAM_reset_sorting_writer { HashTableRemoveAll( writer -> block_no_p1_to_vpos_tab ); }
+
 // MUST be called by THREAD 0!
 void SamBam_writer_sort_bins_to_BAM(SamBam_Writer * writer){
 	SamBam_writer_one_thread_merge_sortedbins(writer);
@@ -2303,65 +2370,51 @@ void SamBam_writer_sort_bins_to_BAM(SamBam_Writer * writer){
 		}
 	}
 
-	HashTable * bin_chunks_table = NULL; // bin_NO => ArrayList [ chunk0_Voff_start, chunk0_Voff_end, chunk1_Voff_start, chunk1_Voff_end, ... ]
 	ArrayList * bins_list = NULL; //	[Bin_Number_0, Bin_Number_1, ...]
 	ArrayList * win16k_list = NULL;	// [ Voffset_0, Voffset_1, ... ]
+	HashTable * bin_chunks_table=NULL;
 	SAMBAM_renew_BAItabs;
 
 	int chro_no = (int)(current_min >> 32);
-	int old_chro_no = -1, last_written_BAI_chro = -1;;
+	int old_chro_no = -1, last_written_BAI_chro = -1;
 
 	int wlen = fwrite("BAI\1", 4, 1, writer -> BAI_fp);
 	if(wlen < 1) assert(wlen >0);
-	int n_ref = writer -> chromosome_name_table -> numOfElements;
+	int thread_i, n_ref = writer -> chromosome_name_table -> numOfElements;
+	subread_lock_t initlocks[writer -> threads];
 	wlen = fwrite(&n_ref, 4, 1, writer -> BAI_fp);
-	SAMBAM_write_empty_chros(0, chro_no)
+	SAMBAM_write_empty_chros(0, chro_no);
 
+	writer -> block_no_p1_to_vpos_tab = HashTableCreate(100000);
+	writer -> writer_threads = calloc(sizeof(struct SamBam_sorted_compressor_st), writer -> threads);
+	for(thread_i = 0; thread_i < writer -> threads; thread_i++){
+		memset(&(writer -> writer_threads[thread_i].strm),0,sizeof(z_stream));
+		deflateInit2(&(writer -> writer_threads[thread_i].strm), writer -> fastest_compression? SAMBAM_COMPRESS_LEVEL_FASTEST : SAMBAM_COMPRESS_LEVEL_NORMAL , Z_DEFLATED,
+			SAMBAM_GZIP_WINDOW_BITS, Z_DEFAULT_MEM_LEVEL, Z_DEFAULT_STRATEGY);
+		subread_init_lock(initlocks+thread_i);
+		subread_lock_occupy(initlocks+thread_i);
+
+		void ** thparam = malloc(sizeof(void*)*3);
+		thparam[0] = writer;
+		thparam[1] = NULL + thread_i;
+		thparam[2] = initlocks+thread_i;
+		pthread_create(&writer -> writer_threads[thread_i].thread_stub, NULL,SamBam_writer_sorted_compress,thparam);
+	}
+	for(thread_i = 0; thread_i < writer -> threads; thread_i++){
+		subread_lock_occupy(initlocks+thread_i);
+		subread_destroy_lock(initlocks+thread_i);
+	}
+
+	SAMBAM_reset_sorting_writer;
 	while(1){ // GO THROUGH EACH READ and write into BAM
-		if(old_chro_no >=0 && (chro_no != old_chro_no || current_min_batch < 0)){ // if there is old_chro && if has to write 
-			if(1){
-				HashTable * new_bin_tab = NULL;
-				ArrayList * new_bin_list = NULL;
-				SamBam_writer_optimize_bins(bin_chunks_table , bins_list,&new_bin_tab,&new_bin_list);
-				bin_chunks_table = new_bin_tab;
-				bins_list = new_bin_list;
-			}
-			int n_bins = bins_list->numOfElements;
-			int win16_nos = win16k_list->numOfElements;
-			//SUBREADprintf("WIN_ITEMS=%d\t\t\tOLD_CHRO=%d  NEW_CHRO=%d  CURR_BATCH=%d\n", win16_nos, old_chro_no, chro_no , current_min_batch);
-
-			fwrite(&n_bins, 4, 1, writer -> BAI_fp);
-			int bini;
-			for(bini = 0; bini < n_bins; bini ++){
-				int bin_no = ArrayListGet(bins_list, bini)-NULL;
-				ArrayList * chunks_list = HashTableGet(bin_chunks_table, NULL+ bin_no+1);
-
-				int n_chunks = chunks_list -> numOfElements/2;
-				fwrite(&bin_no, 4, 1, writer -> BAI_fp);
-				fwrite(&n_chunks, 4, 1, writer -> BAI_fp);
-				int chunk_i;
-				for(chunk_i = 0; chunk_i < 2* n_chunks; chunk_i +=2){
-					long long int Voff_st = ArrayListGet(chunks_list, chunk_i) - NULL;
-					long long int Voff_en = ArrayListGet(chunks_list, chunk_i+1) - NULL;
-					fwrite(&Voff_st, 8, 1, writer -> BAI_fp);
-					fwrite(&Voff_en, 8, 1, writer -> BAI_fp);
-				}
-			}
-
-			fwrite(&win16_nos, 4, 1, writer -> BAI_fp);
-			for(bini = 0; bini < win16_nos; bini++){
-				long long int Voff = ArrayListGet(win16k_list , bini) - NULL;
-				fwrite(&Voff, 8, 1, writer -> BAI_fp);
-			}
+		//if(current_min_batch<0)SUBREADprintf("FINALTEST: OLD/NEW = %d , %d\n", old_chro_no ,chro_no);
+		if(old_chro_no >=0 && chro_no != old_chro_no){ // if there is old_chro && if has to write. Note: old_chro_no<0 means we reached the "unmapped" part of the sorted temp files
+			//SUBREADprintf("\n====== W1CHR %d =====\n", old_chro_no);
+			SamBam_write_BAI_for_1chr(writer, &bin_chunks_table, &bins_list, &win16k_list);
 			last_written_BAI_chro = old_chro_no;
-			SAMBAM_write_empty_chros(old_chro_no +1, chro_no)
-			SAMBAM_renew_BAItabs
-		}
-
-		if(current_min_batch<0 && chro_no < 0 && n_ref >0 && last_written_BAI_chro < n_ref ){
-			int bini;
-			for(bini = last_written_BAI_chro; bini < n_ref; bini++)
-				fwrite("\0\0\0\0\0\0\0\0\0\0\0\0", 4, 2, writer -> BAI_fp);
+			SAMBAM_write_empty_chros(old_chro_no +1, chro_no < 0?n_ref:chro_no);
+			SAMBAM_renew_BAItabs;
+			SAMBAM_reset_sorting_writer;
 		}
 
 		if(current_min_batch>-1){
@@ -2370,7 +2423,11 @@ void SamBam_writer_sort_bins_to_BAM(SamBam_Writer * writer){
 			current_min = SUBREAD_MAX_ULONGLONG;
 			current_min_batch = -1;
 			old_chro_no = chro_no;
-		}else break; // all finished. quit.
+		}else{
+			SamBam_write_sorted_thread_collect(writer);
+			break; // all finished. quit.
+		}
+		
 
 		for(bii = 0; bii < writer -> sorted_batch_id; bii++){
 			if(current_min_fps[bii] < SUBREAD_MAX_ULONGLONG && current_min_fps[bii] < current_min){
@@ -2380,7 +2437,6 @@ void SamBam_writer_sort_bins_to_BAM(SamBam_Writer * writer){
 		}
 		chro_no = (int)(current_min >> 32);
 	}
-	if(writer -> chunk_buffer_used >0)SamBam_writer_add_chunk(writer, -1);
 
 	for(bii = 0; bii < writer -> sorted_batch_id; bii++){
 		if(!sb_fps[bii]) continue;
@@ -2395,16 +2451,34 @@ void SamBam_writer_sort_bins_to_BAM(SamBam_Writer * writer){
 		ArrayListDestroy(bins_list);
 		ArrayListDestroy(win16k_list);
 	}
+
+	terminate_workers(&writer->sorted_notifier);
+	for(thread_i = 0; thread_i < writer -> threads; thread_i++){
+		pthread_join(writer -> writer_threads[thread_i].thread_stub, NULL);
+		deflateEnd(&(writer -> writer_threads[thread_i].strm));
+	}
+	HashTableDestroy(writer -> block_no_p1_to_vpos_tab);
+	
+	free(writer->writer_threads);
 	free(current_min_fps);
 	free(sb_fps);
-}
+} 
 
 void SamBam_writer_finalise_thread(SamBam_Writer * writer, int thread_id){
-	if(writer -> sort_reads_by_coord){
-		SamBam_writer_sort_buff_one_write(writer, writer -> threads_chunk_buffer[thread_id], writer -> threads_chunk_buffer_used[thread_id], thread_id);
-		writer -> threads_chunk_buffer_used [thread_id] = 0;
+	if(writer -> threads < 2){
+		if(writer -> sort_reads_by_coord){
+			SamBam_writer_sort_buff_one_write(writer, writer ->chunk_buffer, writer -> chunk_buffer_used, -1);
+			writer -> chunk_buffer_used= 0;
+		}else{
+			if(writer -> chunk_buffer_used) SamBam_writer_add_chunk(writer, thread_id);
+		}
 	}else{
-		if(writer -> threads_chunk_buffer_used[thread_id]) SamBam_writer_add_chunk(writer, thread_id);
+		if(writer -> sort_reads_by_coord){
+			SamBam_writer_sort_buff_one_write(writer, writer -> threads_chunk_buffer[thread_id], writer -> threads_chunk_buffer_used[thread_id], thread_id);
+			writer -> threads_chunk_buffer_used [thread_id] = 0;
+		}else{
+			if(writer -> threads_chunk_buffer_used[thread_id]) SamBam_writer_add_chunk(writer, thread_id);
+		}
 	}
 }
 
@@ -2421,3 +2495,181 @@ void SamBam_writer_finalise_one_thread(SamBam_Writer * writer){
 		}
 	}
 }
+
+
+unsigned int FC_CRC32(char * dat, int len){
+	unsigned int crc0 = crc32(0, NULL, 0);
+	unsigned int ret = crc32(crc0, (unsigned char *)dat, len);
+	return ret;
+}
+
+void simple_bam_writer_deallocate_index_per_chro(void * p){
+	struct simple_bam_writer_index_per_chro * ch = p;
+	HashTableDestroy(ch->index_binP1_table);
+	ArrayListDestroy(ch->index_binP0_list);
+	ArrayListDestroy(ch->win16k_list);
+	free(ch);
+}
+
+
+
+struct simple_bam_writer_index_per_chro * simple_bam_writer_new_index_per_chro(){
+	struct simple_bam_writer_index_per_chro * ret = malloc(sizeof(struct simple_bam_writer_index_per_chro ));
+	ret -> index_binP1_table = HashTableCreate(4000);
+	HashTableSetDeallocationFunctions(ret -> index_binP1_table, NULL, (void (*) (void*)) ArrayListDestroy);
+	ret -> index_binP0_list = ArrayListCreate(20000);
+	ret -> win16k_list = ArrayListCreate(20000);
+	return ret;
+}
+
+void simple_bam_writer_update_index(simple_bam_writer * writer, char * rbin, int binlen, srInt_64 block_number, int inbin_pos){
+	int chro_no=0;
+	memcpy(&chro_no, rbin + 4, 4);
+	if(chro_no<0)return;
+
+	unsigned int pos, bin_mq_nl=0;
+	pos = *(unsigned int*)(rbin + 8);
+	bin_mq_nl = *(unsigned int*)(rbin + 12);
+
+	struct simple_bam_writer_index_per_chro * index_chro = HashTableGet(writer -> index_per_chro, NULL+chro_no+1);
+	if(NULL==index_chro){
+		index_chro = simple_bam_writer_new_index_per_chro();
+		HashTablePut(writer -> index_per_chro, NULL+chro_no+1, index_chro);
+	}
+
+	unsigned int binno = bin_mq_nl>>16;
+	int cigar_span = SamBam_writer_calc_cigar_span(rbin +4);
+	int this_w16_no = (pos + cigar_span) >>14;      // WIN is calculated on 0-based pos.
+	unsigned long long this_Vpos = block_number<<16 | inbin_pos;
+	ArrayList * win16k_list = index_chro -> win16k_list;
+	// if this read is after the maximum coordinate in the win16k list: all elements before last one and this one starts at this read.
+	if(this_w16_no > win16k_list->numOfElements){
+		int bbi;
+		for(bbi = win16k_list->numOfElements; bbi <=this_w16_no; bbi++)
+			ArrayListPush(win16k_list, NULL+ this_Vpos);
+	}
+
+	ArrayList * this_bin_chunks = HashTableGet(index_chro -> index_binP1_table, NULL+binno+1);
+	if(NULL == this_bin_chunks){
+		this_bin_chunks = ArrayListCreate(4);
+		HashTablePut(index_chro -> index_binP1_table, NULL+binno+1, this_bin_chunks);
+		ArrayListPush(index_chro -> index_binP0_list, NULL+binno);
+	}
+	int found = 0;
+	// a bin is not necessarily continuous. Say, a top-level bin only contains a few reads (most reads a in low-level bins), but their locations are everywhere
+
+	if(this_bin_chunks -> numOfElements > 0){
+		long long diff = this_Vpos >>16;
+		diff -=(this_bin_chunks -> elementList [ this_bin_chunks -> numOfElements - 1] - NULL)>>16;
+		if(diff < MAX_ALLOWED_GAP_IN_BAI_CHUNK){
+			this_bin_chunks -> elementList [ this_bin_chunks -> numOfElements - 1] = NULL+this_Vpos + binlen + 4;
+			found = 1;
+		}
+	}
+	// if the last chunk in this bin isn't good to be extended (too far from the file location of the new read), a new chunk is created.
+	if(!found){
+		ArrayListPush(this_bin_chunks, NULL + this_Vpos);
+		ArrayListPush(this_bin_chunks, NULL + this_Vpos + binlen+4);
+	}
+}
+
+void simple_bam_write_compressed_block(simple_bam_writer * writer,char *obuf, int olen, int ilen, unsigned int crcval, srInt_64 block_number){
+	if(block_number >= 0)HashTablePut(writer -> bam_blockP1_to_offset0B_table, NULL+1+block_number, NULL+ftello(writer -> bam_FP));
+	fwrite("\x1f\x8b\x8\x4\0\0\0\0\0\0\x6\0", 1, 12, writer -> bam_FP);
+	fwrite("\x42\x43\x2\0", 1, 4, writer -> bam_FP);
+
+	int BSIZE = olen+19+6;
+	fwrite(&BSIZE, 1, 2, writer -> bam_FP);
+	fwrite(obuf, 1, olen, writer -> bam_FP);
+	fwrite(&crcval, 1, 4, writer -> bam_FP);
+	fwrite(&ilen, 1, 4, writer -> bam_FP);
+}
+
+void simple_bam_write(void * bin, int binlen, simple_bam_writer * writer, int force_flush){
+	while(binlen > 0 || (force_flush && writer->inbin_len)){
+		int concatinate_binlen = min(binlen, 63000 - writer->inbin_len);
+		memcpy(writer->inbin+writer->inbin_len, bin, concatinate_binlen);
+
+		writer->inbin_len += concatinate_binlen;
+		bin += concatinate_binlen;
+		binlen -= concatinate_binlen;
+		if(writer->inbin_len >=63000 || force_flush){
+			deflateInit2(&writer -> strm, Z_BEST_SPEED, Z_DEFLATED, -15, Z_DEFAULT_MEM_LEVEL, Z_DEFAULT_STRATEGY);
+			char obuf[MERGER_WORKER_BINSIZE];
+			writer->strm.next_in = (unsigned char *)writer->inbin;
+			writer->strm.avail_in = writer->inbin_len;
+			writer->strm.next_out = (unsigned char *)obuf;
+			writer->strm.avail_out = MERGER_WORKER_BINSIZE;
+			deflate(&writer->strm, Z_FINISH);
+			int have = MERGER_WORKER_BINSIZE-writer->strm.avail_out;
+			simple_bam_write_compressed_block(writer, obuf, have, writer->inbin_len, FC_CRC32(writer->inbin, writer->inbin_len), -1);
+			writer->inbin_len=0;
+			deflateEnd(&writer -> strm);
+		}
+	}
+}
+
+simple_bam_writer * simple_bam_create(char * fname){
+	simple_bam_writer * ret = calloc(sizeof(simple_bam_writer), 1);
+	ret -> bam_FP = fopen(fname, "wb");
+	ret -> bam_blockP1_to_offset0B_table = HashTableCreate(100000);
+	simple_bam_write("BAM\1", 4, ret, 0);
+
+	char bainame [strlen(fname)+10];
+	strcpy(bainame , fname);
+	strcat(bainame, ".bai");
+	ret -> bai_FP = fopen(bainame, "wb");
+	fwrite("BAI\1", 1, 4, ret -> bai_FP);
+	ret -> index_per_chro = HashTableCreate(1000);
+	HashTableSetDeallocationFunctions(ret -> index_per_chro , NULL , simple_bam_writer_deallocate_index_per_chro);
+	return ret;
+}
+
+#define vpos_to_rpos rposone = (vposone & 0xffff ) + ( (HashTableGet(writer -> bam_blockP1_to_offset0B_table, NULL+1+(vposone >>16)) - NULL) << 16 )
+
+#define BAM_EOF_MARKER "\x1f\x8b\x08\x04\0\0\0\0\0\xff\x06\0\x42\x43\x02\0\x1b\0\x03\0\0\0\0\0\0\0\0\0"
+void simple_bam_close(simple_bam_writer * writer){
+	fwrite(BAM_EOF_MARKER, 1, 28, writer -> bam_FP);
+	fclose(writer -> bam_FP);
+
+	fwrite(&writer -> total_chromosomes, 1, 4, writer -> bai_FP);
+	int chri;
+	for(chri=0; chri<writer -> total_chromosomes; chri++){
+		struct simple_bam_writer_index_per_chro *this_idx = HashTableGet(writer -> index_per_chro , NULL+1+chri);
+		if(NULL == this_idx ){
+			fwrite("\0\0\0\0\0\0\0\0", 1, 8, writer -> bai_FP);//0 intervals and 0 bins
+		}else{
+			HashTable * new_tab=NULL;
+			ArrayList * new_arr=NULL;
+			SamBam_writer_optimize_bins(this_idx -> index_binP1_table, this_idx -> index_binP0_list ,& new_tab, & new_arr);
+			this_idx -> index_binP1_table = new_tab;
+			this_idx -> index_binP0_list = new_arr;
+			fwrite(&this_idx -> index_binP0_list->numOfElements ,1, 4, writer -> bai_FP);
+			int bini;
+			for(bini = 0; bini < this_idx -> index_binP0_list -> numOfElements; bini ++){
+				int binno = ArrayListGet(this_idx -> index_binP0_list, bini)-NULL;
+				ArrayList * bingaps = HashTableGet(this_idx -> index_binP1_table, NULL+1+binno);
+				srInt_64 gapi = bingaps -> numOfElements/2;
+				fwrite(&binno, 1, 4, writer -> bai_FP);
+				fwrite(&gapi ,1, 4, writer -> bai_FP);
+				for(gapi = 0; gapi < bingaps -> numOfElements; gapi++){
+					srInt_64 rposone, vposone = ArrayListGet(bingaps , gapi)-NULL;
+					vpos_to_rpos;
+					fwrite(&rposone, 1, 8, writer -> bai_FP);
+				}
+			}
+
+			fwrite(&this_idx -> win16k_list -> numOfElements ,1, 4, writer -> bai_FP);
+			for(bini = 0; bini < this_idx -> win16k_list -> numOfElements; bini ++){
+				srInt_64 rposone, vposone = ArrayListGet(this_idx -> win16k_list , bini )-NULL;
+				vpos_to_rpos;
+				fwrite(&rposone, 1, 8, writer -> bai_FP);
+			}
+		}
+	}
+	HashTableDestroy(writer -> index_per_chro);
+	fclose(writer -> bai_FP);
+	free(writer);
+}
+
+
