@@ -33,7 +33,12 @@
 
 #include "hashtable.h" 
 
-#define MAX_SCRNA_FASTQ_FILES 200
+#define SCRNA_HIGHEST_REPORTED_ALIGNMENTS 30
+#define MAX_SCRNA_FASTQ_FILES 256 
+#define SCRNA_FASTA_SPLIT1 "|Rsd:cCounts:mFQs|"
+#define SCRNA_FASTA_SPLIT2 "|Rsd:cCounts:1mFQ|"
+
+
 #define SAM_FLAG_PAIRED_TASK	0x01
 #define SAM_FLAG_FIRST_READ_IN_PAIR 0x40
 #define SAM_FLAG_SECOND_READ_IN_PAIR 0x80
@@ -103,7 +108,9 @@
 #define	IS_BREAKEVEN_READ (8192*4)
 #define IS_R1R2_EQUAL_LEN 1024
 
+#ifndef MAKE_CELLCOUNTS
 #define USE_POSIX_MUTEX_LOCK
+#endif
 
 #if defined(MACOS) || defined(FREEBSD) || defined(USE_POSIX_MUTEX_LOCK)
 typedef pthread_mutex_t subread_lock_t;
@@ -196,6 +203,13 @@ typedef short gene_vote_number_t;
 #define GENE_VOTE_SPACE 173 
 #define GENE_VOTE_TABLE_SIZE 331
 #else
+/*
+#warning "============ ADD '-21' from next line  ================"
+#define GENE_VOTE_SPACE (24 - 21)
+#warning "============ ADD '-13' from next line  ================"
+#define GENE_VOTE_TABLE_SIZE (30 - 13)
+
+*/
 #define GENE_VOTE_SPACE 24
 #define GENE_VOTE_TABLE_SIZE 30 
 #endif
@@ -220,9 +234,12 @@ typedef short gene_vote_number_t;
 //#define base2int(c) ((c)=='A'?0:((c)=='T'?3:((c)=='C'?2:1)))
 #define base2int(c) ((c)<'G'?((c)=='A'?0:2):((c)=='G'?1:3))
 
+/*
+#define base2int(c) (("\x3\x3\x3\x3\x3\x3\x3\x3" "\x3\x3\x3\x3\x3\x3\x3\x3"  "\x3\x3\x3\x3\x3\x3\x3\x3"  "\x3\x3\x3\x3\x3\x3\x3\x3"  "\x3\x3\x3\x3\x3\x3\x3\x3"  "\x3\x3\x3\x3\x3\x3\x3\x3"  "\x3\x3\x3\x3\x3\x3\x3\x3"  "\x3\x3\x3\x3\x3\x3\x3\x3"    "\x3"\
+	 "\x0\x3\x2\x3\x3\x3\x1\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3")[(int)(c)])
+         // A  B  C  D  E  F  G
+*/
 
-                     // A  B  C  D  E  F  G
-//#define base2int(c) ("\x0\x0\x2\x0\x0\x0\x1\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3\x3"[(c)-'A'])
 //#define int2base(c) ((c)==1?'G':((c)==0?'A':((c)==2?'C':'T')))
 //#define int2base(c) ("AGCT"[(c)]) 
 #define int2base(c) (1413695297 >> (8*(c))&0xff)
@@ -301,9 +318,9 @@ typedef struct{
 	void * appendix2;
 } gene_value_index_t;
 
-
 typedef struct {
 	gene_vote_number_t max_vote;
+	int max_vote_IJ;
 	gehash_data_t max_position;
 	gene_quality_score_t max_quality;
 	gene_vote_number_t max_indel_recorder[MAX_INDEL_TOLERANCE*3];
@@ -321,6 +338,8 @@ typedef struct {
 	gene_vote_number_t indel_recorder [GENE_VOTE_TABLE_SIZE][GENE_VOTE_SPACE][MAX_INDEL_TOLERANCE*3];
 	char current_indel_cursor[GENE_VOTE_TABLE_SIZE][GENE_VOTE_SPACE];
 	char toli[GENE_VOTE_TABLE_SIZE][GENE_VOTE_SPACE];
+	int topK_votes[SCRNA_HIGHEST_REPORTED_ALIGNMENTS];
+	int topK_IJ[SCRNA_HIGHEST_REPORTED_ALIGNMENTS ];
 
 	#ifdef MAKE_FOR_EXON
 	short coverage_start [GENE_VOTE_TABLE_SIZE][GENE_VOTE_SPACE];
@@ -395,7 +414,8 @@ struct thread_input_buffer {
 
 #define SEEKGZ_CHAIN_BLOCKS_NO 15
 #define SEEKGZ_ZLIB_WINDOW_SIZE (32*1024)
-
+#define PAIRER_GZIP_WINDOW_BITS -15
+#define PAIRER_DEFAULT_MEM_LEVEL 8
 
 typedef struct {
 	unsigned long long block_start_in_file_offset;
@@ -476,16 +496,16 @@ typedef struct {
 	int is_EOF;
 } input_BLC_t;
 
+#define MAX_SCRNA_READ_LENGTH 160 
+
 typedef struct{
-	union{
-		srInt_64 pos_file1, pos_file2, pos_file3;
-		seekable_position_t zpos_file1;
-	};
-	seekable_position_t zpos_file2;
-	seekable_position_t zpos_file3;
-	int current_file_no;
-	srInt_64 current_read_no;
-} input_mFQ_pos_t;
+	char read_text[MAX_SCRNA_READ_LENGTH];
+	char qual_text[MAX_SCRNA_READ_LENGTH];
+	int read_len;
+	int is_last_read_in_dataset;
+	int file_no;
+	int guessed_lane_no;
+} input_mFQ_cached_read_t;
 
 typedef struct {
 	char filename[MAX_FILE_NAME_LENGTH+1];
@@ -493,24 +513,82 @@ typedef struct {
 	int is_plain;
 	FILE * plain_fp;
 	seekable_zfile_t gz_fp;
+	gzFile zlib_fp;
 	int is_first_chars;
 	unsigned char first_chars[2];
 } autozip_fp;
 
+/*
+typedef pthread_mutex_t  cellCounts_lock_t;
+#define cellCounts_init_lock(pp) pthread_mutex_init(pp, NULL)
+#define cellCounts_destroy_lock pthread_mutex_destroy
+#define cellCounts_lock_occupy pthread_mutex_lock
+#define cellCounts_lock_release pthread_mutex_unlock
+*/
+
+typedef struct{
+	int is_spin_lock;
+	pthread_spinlock_t spinlock;
+	pthread_mutex_t mutexlock;
+} cellCounts_lock_t;
+
+void cellCounts_init_lock(cellCounts_lock_t * lock, int is_spin);
+void cellCounts_destroy_lock(cellCounts_lock_t * lock);
+int cellCounts_lock_occupy(cellCounts_lock_t * lock);
+int cellCounts_lock_release(cellCounts_lock_t * lock);
+
+/*
+#define cellCounts_init_lock(pp) pthread_spin_init(pp,PTHREAD_PROCESS_SHARED)
+#define cellCounts_destroy_lock pthread_spin_destroy 
+#define cellCounts_lock_occupy pthread_spin_lock
+#define cellCounts_lock_release pthread_spin_unlock
+*/
 
 
 typedef struct {
-	char ** files1;
-	char ** files2;
-	char ** files3;
-	int total_files;
-	int current_file_no;
-	srInt_64 current_read_no;
-	autozip_fp autofp1;
-	autozip_fp autofp2;
-	autozip_fp autofp3;
+        char ** files1;
+        char ** files2;
+        char ** files3;
+        int total_files;
+        int current_file_no;
+        int current_guessed_lane_no;
+        srInt_64 current_read_no;
+        autozip_fp autofp1;
+        autozip_fp autofp2;
+        autozip_fp autofp3;
 } input_mFQ_t;
 
+
+typedef struct
+{
+        char chro_name[MAX_CHROMOSOME_NAME_LEN];
+        unsigned int chro_length;
+} SamBam_Reference_Info;
+
+
+typedef struct{
+	int current_BAM_file_no;
+	srInt_64 section_start_pos;
+	int in_section_offset;
+	srInt_64 current_read_no;
+} input_scBAM_pos_t;
+
+typedef struct {
+	FILE * os_file;
+	char * BAM_file_names[MAX_SCRNA_FASTQ_FILES];
+	char section_buff[66000];
+	char align_buff[FC_LONG_READ_RECORD_HARDLIMIT];
+	int current_BAM_file_no;
+	int total_BAM_files;
+	int in_section_offset;
+	int section_bin_bytes;
+	int chro_table_size;
+	SamBam_Reference_Info * chro_table;
+	srInt_64 section_start_pos;
+	srInt_64 current_read_no;
+	subread_lock_t read_lock;
+} input_scBAM_t;
+
 typedef struct {
 	int read_no_in_chunk;
 	int reads_available_in_chunk; // -1 : EOF of all input reads : no next chunk available.
@@ -538,11 +616,26 @@ typedef struct {
 	int is_EOF;
 } cache_BCL_t;
 
+
+typedef struct{
+        union{
+                srInt_64 pos_file1, pos_file2, pos_file3;
+                seekable_position_t zpos_file1;
+        };
+        seekable_position_t zpos_file2;
+        seekable_position_t zpos_file3;
+        int current_file_no;
+        srInt_64 current_read_no;
+} input_mFQ_pos_t;
+
 typedef struct{
 	union{
 		unsigned long long simple_file_position;
 		seekable_position_t seekable_gzip_position;
 		input_BLC_pos_t BCL_position;
+		input_scBAM_pos_t scBAM_position;
+                input_mFQ_pos_t mFQ_position;
+
 	};
 	char gzfa_last_name[MAX_READ_NAME_LEN];
 } gene_inputfile_position_t;
@@ -595,6 +688,7 @@ typedef struct {
 	union{
 		cache_BCL_t bcl_input;
 		input_mFQ_t scRNA_fq_input;
+		input_scBAM_t scBAM_input;
 	};
 } gene_input_t;
 
